//===========================================================
//=     University of Illinois at Urbana/Champaign          =
//=     Department of Computer Science                      =
//=     Dr. Dan Roth - Cognitive Computation Group          =
//=                                                         =
//=  Project: Snow                                          =
//=                                                         =
//=   Module: Winnow.cpp                                    =
//=  Version: 2.1                                           =
//=  Authors: Jeff Rosen and Andrew Carlson                 =
//=     Date: xx/xx/99                                      = 
//=                                                         =
//===========================================================

#include "Winnow.h"
#include "Target.h"
#include "GlobalParams.h"
#include <math.h>
#include <iomanip>
#include <vector>

// Presents an example to the Categories classifier.  The return
// value indicates whether the Categories classifier made a mistake
// on this example (and had its weight vector updated).
// Return of true indicates a mistake and update, otherwise false.

bool Winnow::PresentExample(  Target& tar, Example& ex )
{
    bool result = false;
    
    // Reset activation flags
    tar.externalActivation = ex.HasFeature(tar.targetID);
    tar.internalActivation = 0.0;
    
    Example::FeatureMap::const_iterator it = ex.GetFeatureBegin();
    Example::FeatureMap::const_iterator end = ex.GetFeatureEnd();

    Target::FeatureMap::iterator f;
    
    double totalWeight = 0.0;
    
    for ( ; it != end; it++)
    {
        // Don't include this target
        if ((it->first != tar.targetID) && ((!globalParams.multipleLabels) ||
         (globalParams.targetIds.find(it->first) ==
            globalParams.targetIds.end())))
        {
            // Does the feature already exist?
            if (((f = tar.features.find(it->first)) != tar.features.end()) &&
                (f->second.discarded == active))
            {
                if ((globalParams.currentCycle == 1) &&
                    (tar.externalActivation))
                {
                    f->second.activeCount++;
                }
                totalWeight += it->second * f->second.weight;
            } else {
                // either feature has not been seen or hasn't
                // met the eligibility threshold yet
               
                int actualCount = 1;
                
                if (tar.externalActivation)
                {
                    Target::FeatureMap::iterator fc;
                    if ((fc = tar.features.find(it->first))
                        != tar.features.end())
                    {
                        if (tar.externalActivation 
                            && (globalParams.currentCycle == 1)) 
                        {
                            fc->second.activeCount++;
                        }   
                        actualCount = fc->second.activeCount;
                    } else {
                        // Add the feature
                        tar.features.insert(
                            make_pair(it->first, MinimalFeature()));
                        tar.features[it->first].discarded = pending;
                        if (tar.externalActivation) 
                        {
                            tar.features[it->first].activeCount = 1;
                        }
                    }
                }
                
                // Check the eligibility of the feature
                if ((actualCount >= globalParams.eligibilityThreshold)
                    && ((globalParams.sparseNetwork && tar.externalActivation)
                        || !globalParams.sparseNetwork))
                {
                    if (globalParams.verbosity >= VERBOSE_MAX)
                        cout << "Winnow adding feature (" << it->first
                             << ") to target (" << tar.targetID
                             << ")" << endl;

                    // If it is eligible, set its weight
                    // (and increment the totalWeight)
                    if (tar.externalActivation)
                    {
                        tar.features[it->first].discarded = active;
                        tar.activeFeatureCount++;
                    }
                    
                    tar.features[it->first].weight = defaultWeight;
                    //totalWeight += defaultWeight;
                }
                
            }
        } 
    }

    // Update the active / non-active counters
    if (globalParams.currentCycle == 1)
       if (tar.externalActivation) 
          tar.activeCount++; 
       else 
          tar.nonActiveCount++;

    // Update internal activation, if necessary
    tar.internalActivation = totalWeight / threshold;
    bool intActivation = (tar.internalActivation >= 1.0);

    // If there is a mistake let the algorithm update the weights
    if ((tar.externalActivation != intActivation)
        && ((globalParams.currentCycle > 1)
            || !globalParams.noFirstCycleUpdate))
    {
        if (globalParams.verbosity >= VERBOSE_MED)
            cout << "Updating target (" << tar.targetID
                 << ") for mistake..." << endl;
        
        Update(tar, ex);

        result = true;
    }

    return result;
}

// Updates the target based on the example and the Winnow parameters

void Winnow::Update( Target& tar, Example& ex )
{
    // Assume false negative
    double multiplier = alpha;

    // Update only called if a mistake was made.
    // If the target was internally activated, then the mistake
    // must have been a false positive.
    if (tar.internalActivation >= 1.0)  multiplier = beta;

    // Declare the iterators we will use to access the intersection
    // of the examples active features with the categories weight
    // vector.
    Example::FeatureMap::const_iterator it = ex.GetFeatureBegin();
    Example::FeatureMap::const_iterator end = ex.GetFeatureEnd();

    Target::FeatureMap::iterator f;

    // Update the weight vector
    for ( ; it != end; it++)
    {
        if ((f = tar.features.find(it->first)) != tar.features.end()) 
        {
           if (tar.features[it->first].discarded == active)
           {
              f->second.weight *= multiplier;
              f->second.updates++; 
           }
        }
    }

    // Update the confidence of the target
    double x = sqrt((double)(tar.activeCount + tar.nonActiveCount));
    tar.confidence /= x / (((x - 1) / alpha) + 1);
}

void Winnow::Discard( Target& tar )
{
    Target::FeatureMap::iterator feat = tar.features.begin();

    // Update the weight vector
    for ( ; feat != tar.features.end(); feat++)
    {
        if ((globalParams.discardMethod == DISCARD_ABS)
            && (feat->second.weight < globalParams.discardThreshold))
        {
            feat->second.discarded = discard;
            tar.activeFeatureCount--;
        }
    }
}

void Winnow::SetTargetActivation( Target& tar, Example& ex )
{
    Example::FeatureMap::const_iterator it = ex.GetFeatureBegin();
    Example::FeatureMap::const_iterator end = ex.GetFeatureEnd();

    Target::FeatureMap::iterator f;
    
    tar.externalActivation = ex.HasFeature(tar.targetID);
    if (tar.externalActivation) tar.activeCount++; else tar.nonActiveCount++;
    tar.internalActivation = 0.0;

    for ( ; it != end; it++)
    {
        // Don't process the target itself
        if (it->first != tar.targetID)
        {
            // Does this feature have a weight?
            if (((f = tar.features.find(it->first)) != tar.features.end()) 
                && (f->second.discarded == active)) 
            {
                tar.internalActivation += it->second * f->second.weight;
            } else {
                // Don't smooth any targets!
                if (globalParams.targetIds.find(it->first) ==
                    globalParams.targetIds.end())
                    tar.internalActivation -= it->second * globalParams.smoothing;
            } 
        }
    }
}

double Winnow::ReturnNormalizedActivation( Target& tar)
{
    double normalizedActivation = 
        (1.0 / (1.0 + exp(-(tar.InternalActivation() -
                                  threshold) )) ) * tar.Confidence();
    return normalizedActivation;
}

void Winnow::Show( ostream& out )
{
/*
#ifdef WIN32
    operator<<(out, "Winnow network:(");
    out << setprecision(4) << alpha;
    operator<<(out, ", ");
    out << setprecision(4) << beta;
    operator<<(out, ", ");
    out << setprecision(4) << threshold;
    operator<<(out, ", ");
    out << setprecision(4) << defaultWeight << ')' << endl;
#else
*/
    out << "Winnow network:("
        << setprecision(4) << alpha << ", "
        << setprecision(4) << beta << ", "
        << setprecision(4) << threshold << ", "
        << setprecision(4) << defaultWeight << ")" << endl;
//#endif
}

void Winnow::Read( ifstream& in )
{
    in >> alpha >> beta >> threshold >> defaultWeight;
    
    if (globalParams.verbosity >= VERBOSE_MED)
    {
        cout << endl << "Created:";
        Show(cout);
    }
}

void Winnow::Write( ofstream& out )
{
    if (globalParams.verbosity >= VERBOSE_MED)
    {
        Show(cout);
    }

#ifdef WIN32
    // This looks weird, but it is a work around for the
    // NT version.  Due to a bug in MSVC 5.0 you have to
    // call the insertion operator explicitly.
    operator<<(out, "winnow");
    
    out << ' '
        << index << ' '
        << alpha << ' '
        << beta << ' ' 
        << threshold << ' '
        << defaultWeight << endl;
#else
    out << "winnow "
        << setprecision(10)
        << index << ' '
        << alpha << ' '
        << beta << ' '
        << threshold << ' '
        << defaultWeight << endl;
#endif
}
